r""" This module implements Geometry Score (GS) in PyTorch.
Implementation is inspired by Valentin Khrulkov's (@KhrulkovV) implementation:
https://github.com/KhrulkovV/geometry-score
See paper for details:
https://arxiv.org/pdf/1802.02664.pdf
"""
from typing import Optional, Tuple
from multiprocessing import Pool

from scipy.spatial.distance import cdist

import torch
import numpy as np

from piq.base import BaseFeatureMetric


def relative(intervals: np.ndarray, alpha_max: float, i_max: int = 100) -> np.ndarray:
    r"""
    For a collection of intervals this functions computes
    RLT by formulas (2) and (3) from the paper. This function will be typically called
    on the output of the gudhi persistence_intervals_in_dimension function.
    Args:
      intervals: list of intervals e.g. [[0, 1], [0, 2], [0, np.inf]].
      alpha_max: The maximal persistence value
      i_max: Upper bound on the value of beta_1 to compute.
    Returns:
        rlt: Array of size (i_max, ) containing desired RLT.
    """

    persistence_intervals = []
    # If for some interval we have that it persisted up to np.inf
    # we replace this point with alpha_max.
    for interval in intervals:
        if np.isinf(interval[1]):
            persistence_intervals.append([interval[0], alpha_max])
        else:
            persistence_intervals.append(list(interval))

    # If there are no intervals in H1 then we always observed 0 holes.
    if len(persistence_intervals) == 0:
        rlt = np.zeros(i_max)
        rlt[0] = 1.0
        return rlt

    persistence_intervals_ext = persistence_intervals + [[0, alpha_max]]
    persistence_intervals_ext = np.array(persistence_intervals_ext)
    persistence_intervals = np.array(persistence_intervals)

    # Change in the value of beta_1 may happen only at the boundary points
    # of the intervals
    switch_points = np.sort(np.unique(persistence_intervals_ext.flatten()))
    rlt = np.zeros(i_max)
    for i in range(switch_points.shape[0] - 1):
        midpoint = (switch_points[i] + switch_points[i + 1]) / 2
        s = 0
        for interval in persistence_intervals:
            # Count how many intervals contain midpoint
            if midpoint >= interval[0] and midpoint < interval[1]:
                s = s + 1
        if (s < i_max):
            rlt[s] += (switch_points[i + 1] - switch_points[i])

    return rlt / alpha_max


def lmrk_table(witnesses: np.ndarray, landmarks: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    r"""Construct an input for the gudhi.WitnessComplex function.
    Args:
        witnesses: Array of size w x d, containing witnesses
        landmarks: Array of size l x d, containing landmarks
    Returns:
        distances: 3D array of size w x l x 2. It satisfies the property that
            distances[i, :, :] is [idx_i, dists_i], where dists_i are the sorted distances
            from the i-th witness to each point in L and idx_i are the indices of the corresponding points
            in L, e.g., D[i, :, :] = [[0, 0.1], [1, 0.2], [3, 0.3], [2, 0.4]]
        max_dist: Maximal distance between W and L
    """
    a = cdist(witnesses, landmarks)
    max_dist = np.max(a)
    idx = np.argsort(a)
    b = a[np.arange(np.shape(a)[0])[:, np.newaxis], idx]
    distances = np.dstack([idx, b])
    return distances, max_dist


def witness(
        features: np.ndarray, sample_size: int = 64, gamma: Optional[float] = None) \
        -> Tuple[np.ndarray, np.ndarray]:
    """Compute the persistence intervals for the dataset of features using the witness complex.

    Args:
        features: Array of shape (N_samples, data_dim) representing the dataset.
        sample_size: Number of landmarks to use on each iteration.
        gamma: Parameter determining maximum persistence value. Default is `1.0 / 128 * N_imgs / 5000`

    Returns
        A list of persistence intervals and the maximal persistence value.
    """
    # Install gudhi only if needed
    try:
        import gudhi
    except ImportError as e:
        import six
        error = e.__class__(
            "You are likely missing your GUDHI installation, "
            "you should visit http://gudhi.gforge.inria.fr/python/latest/installation.html "
            "for further instructions.\nIf you use conda, you can use\nconda install -c conda-forge gudhi"
        )
        six.raise_from(error, e)

    N = features.shape[0]
    if gamma is None:
        gamma = 1.0 / 128 * N / 5000

    # Randomly sample `sample_size` points from X
    np.random.seed()
    idx = np.random.choice(N, sample_size)
    landmarks = features[idx]

    distances, max_dist = lmrk_table(witnesses=features, landmarks=landmarks)
    wc = gudhi.WitnessComplex(distances)
    alpha_max = max_dist * gamma
    st = wc.create_simplex_tree(max_alpha_square=alpha_max, limit_dimension=2)

    # This seems to modify the st object
    st.persistence(homology_coeff_field=2)
    intervals = st.persistence_intervals_in_dimension(1)
    return intervals, alpha_max


class GS(BaseFeatureMetric):
    r"""Interface of Geometry Score.
    It's computed for a whole set of data and can use features from encoder instead of images itself to decrease
    computation cost. GS can compare two data distributions with different number of samples.
    But dimensionalities should match, otherwise it won't be possible to correctly compute statistics.

    Args:
        predicted_features (torch.Tensor): Low-dimension representation of predicted image set.
            Shape (N_pred, encoder_dim)
        target_features (torch.Tensor): Low-dimension representation of target image set.
            Shape (N_targ, encoder_dim)

    Returns:
        score (torch.Tensor): Scalar value of the distance between image sets.

    References:
        .. [1] Khrulkov V., Oseledets I. (2018).
        Geometry score: A method for comparing generative adversarial networks.
        arXiv preprint, 2018.
        https://arxiv.org/abs/1802.02664
    """
    def __init__(self, sample_size: int = 64, num_iters: int = 1000, gamma: Optional[float] = None,
                 i_max: int = 100, num_workers: int = 4) -> None:
        r"""
        Args:
            sample_size: Number of landmarks to use on each iteration.
                Higher values can give better accuracy, but increase computation cost.
            num_iters: Number of iterations.
                Higher values can reduce variance, but increase computation cost.
            gamma: Parameter determining maximum persistence value. Default is `1.0 / 128 * N_imgs / 5000`
            i_max: Upper bound on i in RLT(i, 1, X, L)
            num_workers: Number of proccess used for GS computation.


        """
        super().__init__()
        self.sample_size = sample_size
        self.num_iters = num_iters
        self.gamma = gamma
        self.i_max = i_max
        self.num_workers = num_workers

    def compute_metric(self, predicted_features: torch.Tensor, target_features: torch.Tensor) -> torch.Tensor:
        r"""Implements Algorithm 2 from the paper.
        Args:
            predicted_features: Samples from data distribution.
                Shape (N_samples, data_dim).
            target_features: Samples from data distribution.
                Shape (N_samples, data_dim).
        Returns:
            score: Scalar value of the distance between distributions.
        """
        with Pool(self.num_workers) as p:
            self.features = predicted_features.detach().cpu().numpy()
            pool_results = p.map(self._relative_living_times, range(self.num_iters))
            mean_rlt_predicted = np.vstack(pool_results).mean(axis=0)

            self.features = target_features.detach().cpu().numpy()
            pool_results = p.map(self._relative_living_times, range(self.num_iters))
            mean_rlt_target = np.vstack(pool_results).mean(axis=0)

        score = np.sum((mean_rlt_predicted - mean_rlt_target) ** 2)

        return torch.tensor(score, device=predicted_features.device) * 1000

    def _relative_living_times(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Implements Algorithm 1 for two samples of landmarks.
    
        Args:
            idx : Dummy argument. Used for multiprocessing.Pool to work correctly
        
        Returns:
            An array of size (i_max, ) containing RLT(i, 1, X, L)
            for randomly sampled landmarks.
        """
        intervals, alpha_max = witness(
            self.features, sample_size=self.sample_size, gamma=self.gamma)
        rlt = relative(intervals, alpha_max, i_max=self.i_max)

        return rlt



"""
PyTorch implementation of Gradient Magnitude Similarity Deviation (GMSD)
and Multi-Scale Gradient Magnitude Similarity Deviation (MS-GMSD)
Reference:
    Wufeng Xue et al. Gradient Magnitude Similarity Deviation (2013)
    https://arxiv.org/pdf/1308.3052.pdf
    GRADIENT MAGNITUDE SIMILARITY DEVIATION ON MULTIPLE SCALES (2017)
    http://www.cse.ust.hk/~psander/docs/gradsim.pdf

"""
import torch
from typing import Optional, Union, Tuple, List

import torch.nn.functional as F
from torch.nn.modules.loss import _Loss

from piq.utils import _adjust_dimensions, _validate_input
from piq.functional import similarity_map, gradient_map, prewitt_filter


def _gmsd(prediction: torch.Tensor, target: torch.Tensor,
          reduction: Optional[str] = 'mean') -> torch.Tensor:
    r"""Compute Gradient Magnitude Similarity Deviation
    Both inputs supposed to be in range [0, 1] with RGB order.
    Args:
        prediction: Tensor of shape :math:`(N, C, H, W)` holding an distorted image.
        target: Tensor of shape :math:`(N, C, H, W)` holding an target image
        reduction: Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed.

    Returns:
        gmsd : Gradient Magnitude Similarity Deviation between given tensors.

    References:
        https://arxiv.org/pdf/1308.3052.pdf
    """
    # Constant for numerical stability
    EPS: float = 0.0026

    # Convert RGB image to YCbCr and take luminance: Y = 0.299 R + 0.587 G + 0.114 B
    num_channels = prediction.size(1)
    if num_channels == 3:
        prediction = 0.299 * prediction[:, 0, :, :] + 0.587 * prediction[:, 1, :, :] + 0.114 * prediction[:, 2, :, :]
        target = 0.299 * target[:, 0, :, :] + 0.587 * target[:, 1, :, :] + 0.114 * target[:, 2, :, :]

        # Add channel dimension
        prediction = prediction[:, None, :, :]
        target = target[:, None, :, :]
    
    # Compute grad direction
    kernels = torch.stack([prewitt_filter(), prewitt_filter().transpose(-1, -2)])
    pred_grad = gradient_map(prediction, kernels)
    trgt_grad = gradient_map(target, kernels)

    # Compute GMS
    gms = similarity_map(pred_grad, trgt_grad, EPS)
    mean_gms = torch.mean(gms, dim=[1, 2, 3], keepdims=True)

    # Compute GMSD along spatial dimensions. Shape (batch_size )
    score = torch.pow(gms - mean_gms, 2).mean(dim=[1, 2, 3]).sqrt()
    
    if reduction == 'none':
        return score

    return {'mean': score.mean,
            'sum': score.sum
            }[reduction](dim=0)


class GMSDLoss(_Loss):
    r"""Creates a criterion that measures Gradient Magnitude Similarity Deviation
    between each element in the input and target.

    Args:
        reduction: Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed. Default: ``'mean'``
        data_range: The difference between the maximum and minimum of the pixel value,
            i.e., if for image x it holds min(x) = 0 and max(x) = 1, then data_range = 1.
            The pixel value interval of both input and output should remain the same.
            
    Reference:
        Wufeng Xue et al. Gradient Magnitude Similarity Deviation (2013)
        https://arxiv.org/pdf/1308.3052.pdf
        
    """

    def __init__(self, reduction: str = 'mean', data_range: Union[int, float] = 1.) -> None:
        super().__init__()

        # Generic loss parameters.
        self.reduction = reduction

        # Loss-specific parameters.
        self.data_range = data_range

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        r"""Computation of Gradient Magnitude Similarity Deviation (GMSD) as a loss function.

        Args:
            prediction: Tensor of prediction of the network.
            target: Reference tensor.

        Returns:
            Value of GMSD loss to be minimized. 0 <= GMSD loss <= 1.
        """
        _validate_input(input_tensors=(prediction, target), allow_5d=False, scale_weights=None)
        prediction, target = _adjust_dimensions(input_tensors=(prediction, target))

        return self.compute_metric(prediction, target)
    
    def compute_metric(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        
        if self.data_range == 255:
            prediction = prediction / 255.
            target = target / 255.

        # Average by 2x2 filter and downsample
        padding = (prediction.shape[2] % 2, prediction.shape[3] % 2)
        prediction = F.avg_pool2d(prediction, kernel_size=2, stride=2, padding=padding)
        target = F.avg_pool2d(target, kernel_size=2, stride=2, padding=padding)
        
        score = _gmsd(
            prediction, target, reduction=self.reduction)

        return score
    

class MultiScaleGMSDLoss(_Loss):
    r"""Creates a criterion that measures multi scale Gradient Magnitude Similarity Deviation
    between each element in the input :math:`x` and target :math:`y`.


    Args:
        reduction: Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed. Default: ``'mean'``
        data_range: The difference between the maximum and minimum of the pixel value,
            i.e., if for image x it holds min(x) = 0 and max(x) = 1, then data_range = 1.
            The pixel value interval of both input and output should remain the same.
        scale_weights: Weights for different scales. Can contain any number of floating point values.
        chromatic: Flag to use MS-GMSDc algorithm from paper.
            It also evaluates chromatic components of the image. Default: True
        beta1: Algorithm parameter. Weight of chromatic component in the loss.
        beta2: Algorithm parameter. Small constant, see [1].
        beta3: Algorithm parameter. Small constant, see [1].
        
    Reference:
        [1] GRADIENT MAGNITUDE SIMILARITY DEVIATION ON MULTIPLE SCALES (2017)
            http://www.cse.ust.hk/~psander/docs/gradsim.pdf
    """

    def __init__(self, reduction: str = 'mean', data_range: Union[int, float] = 1.,
                 scale_weights: Optional[Union[Tuple[float], List[float]]] = None,
                 chromatic: bool = False, beta1: float = 0.01, beta2: float = 0.32,
                 beta3: float = 15.) -> None:
        super().__init__()

        # Generic loss parameters.
        self.reduction = reduction

        # Loss-specific parameters.
        self.data_range = data_range
        
        # Values from the paper
        if scale_weights is None:
            self.scale_weights = torch.tensor([0.096, 0.596, 0.289, 0.019])
        else:
            # Normalize scale weights
            self.scale_weights = torch.tensor(scale_weights) / torch.tensor(scale_weights).sum()

        self.chromatic = chromatic
        self.beta1 = beta1
        self.beta2 = beta2
        self.beta3 = beta3
            
    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        r"""Computation of Multi scale GMSD as a loss function.

        Args:
            prediction: Tensor of prediction of the network.
            target: Reference tensor.

        Returns:
            Value of GMSD loss to be minimized. 0 <= GMSD loss <= 1.
        """
        _validate_input(input_tensors=(prediction, target), allow_5d=False, scale_weights=self.scale_weights)
        prediction, target = _adjust_dimensions(input_tensors=(prediction, target))

        return self.compute_metric(prediction, target)
    
    def compute_metric(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        # Check that input is big enough
        num_scales = self.scale_weights.size(0)
        min_size = 2 ** num_scales + 1

        if prediction.size(-1) < min_size or prediction.size(-2) < min_size:
            raise ValueError(f'Invalid size of the input images, expected at least {min_size}x{min_size}.')
        
        if self.data_range == 255:
            prediction = prediction / 255.
            target = target / 255.
        
        scale_weights = self.scale_weights.to(prediction)
        ms_gmds = []
        for scale in range(num_scales):
            if scale > 0:
                # Average by 2x2 filter and downsample
                padding = (prediction.shape[2] % 2, prediction.shape[3] % 2)
                prediction = F.avg_pool2d(prediction, kernel_size=2, padding=padding)
                target = F.avg_pool2d(target, kernel_size=2, padding=padding)
        
            score = _gmsd(prediction, target, reduction='none')
            ms_gmds.append(score)
        
        # Stack results in different scales and multiply by weight
        ms_gmds_val = scale_weights.view(1, num_scales) * (torch.stack(ms_gmds, dim=1) ** 2)
  
        # Sum and take sqrt per-image
        ms_gmds_val = torch.sqrt(torch.sum(ms_gmds_val, dim=1))
        
        # Shape: (batch_size, )
        score = ms_gmds_val
        
        if self.chromatic:
            assert prediction.size(1) == 3, "Chromatic component can be computed only for RGB images!"
            
            # Convert to YIQ color space https://en.wikipedia.org/wiki/YIQ
            iq_weights = torch.tensor([[0.5959, -0.2746, -0.3213], [0.2115, -0.5227, 0.3112]]).t().to(prediction)
            prediction_iq = torch.matmul(prediction.permute(0, 2, 3, 1), iq_weights).permute(0, 3, 1, 2)
            target_iq = torch.matmul(target.permute(0, 2, 3, 1), iq_weights).permute(0, 3, 1, 2)
            
            rmse_iq = torch.sqrt(torch.mean((prediction_iq - target_iq) ** 2, dim=[2, 3]))
            rmse_chrome = torch.sqrt(torch.sum(rmse_iq ** 2, dim=1))
            gamma = 2 / (1 + self.beta2 * torch.exp(-self.beta3 * ms_gmds_val)) - 1
            
            score = gamma * ms_gmds_val + (1 - gamma) * self.beta1 * rmse_chrome
            
        if self.reduction == 'none':
            return score

        return {'mean': score.mean,
                'sum': score.sum
                }[self.reduction](dim=0)



__version__ = "0.5.0"

from .ssim import ssim, multi_scale_ssim, SSIMLoss, MultiScaleSSIMLoss
from .msid import MSID
from .fid import FID
from .kid import KID
from .tv import TVLoss, total_variation
from .gmsd import GMSDLoss, MultiScaleGMSDLoss
from .gs import GS
from .isc import IS, inception_score
from .vif import VIFLoss, vif_p
from .brisque import BRISQUELoss, brisque
from .perceptual import StyleLoss, ContentLoss, LPIPS, DISTS
from .psnr import psnr
from .fsim import fsim, FSIMLoss
from .vsi import vsi, VSILoss



import torch
import pytest

from piq import GMSDLoss, MultiScaleGMSDLoss

LEAF_VARIABLE_ERROR_MESSAGE = 'Expected non None gradient of leaf variable'


@pytest.fixture(scope='module')
def prediction() -> torch.Tensor:
    return torch.rand(2, 3, 128, 128)


@pytest.fixture(scope='module')
def target() -> torch.Tensor:
    return torch.rand(2, 3, 128, 128)


# ================== Test class: `GMSDLoss` ==================
def test_gmsd_loss(prediction: torch.Tensor, target: torch.Tensor) -> None:
    loss = GMSDLoss()
    loss(prediction, target)


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_gmsd_loss_on_gpu(prediction: torch.Tensor, target: torch.Tensor) -> None:
    loss = GMSDLoss()
    loss(prediction.cuda(), target.cuda())


def test_gmsd_loss_backward(prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction.requires_grad_()
    loss_value = GMSDLoss()(prediction, target)
    loss_value.backward()
    assert prediction.grad is not None, LEAF_VARIABLE_ERROR_MESSAGE


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_gmsd_loss_backward_on_gpu(prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction.requires_grad_()
    loss_value = GMSDLoss()(prediction.cuda(), target.cuda())
    loss_value.backward()
    assert prediction.grad is not None, LEAF_VARIABLE_ERROR_MESSAGE


def test_gmsd_zero_for_equal_tensors(prediction: torch.Tensor):
    loss = GMSDLoss()
    target = prediction.clone()
    measure = loss(prediction, target)
    assert measure.abs() <= 1e-6, f'GMSD for equal tensors must be 0, got {measure}'


def test_gmsd_loss_raises_if_tensors_have_different_types(target: torch.Tensor) -> None:
    wrong_type_prediction = list(range(10))
    with pytest.raises(AssertionError):
        GMSDLoss()(wrong_type_prediction, target)


def test_gmsd_loss_supports_different_data_ranges(prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction_255 = (prediction * 255).type(torch.uint8)
    target_255 = (target * 255).type(torch.uint8)
    loss = GMSDLoss()
    measure = loss(prediction, target)

    loss_255 = GMSDLoss(data_range=255)
    measure_255 = loss_255(prediction_255, target_255)
    diff = torch.abs(measure_255 - measure)
    assert diff <= 1e-4, f'Result for same tensor with different data_range should be the same, got {diff}'


def test_gmsd_supports_greyscale_tensors():
    loss = GMSDLoss()
    target = torch.ones(2, 1, 128, 128)
    prediction = torch.zeros(2, 1, 128, 128)
    loss(prediction, target)


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_gmsd_supports_greyscale_tensors_on_gpu():
    loss = GMSDLoss()
    target = torch.ones(2, 1, 128, 128).cuda()
    prediction = torch.zeros(2, 1, 128, 128).cuda()
    loss(prediction, target)


# ================== Test class: `MultiScaleGMSDLoss` ==================
def test_multi_scale_gmsd_loss(prediction: torch.Tensor, target: torch.Tensor) -> None:
    loss = MultiScaleGMSDLoss(chromatic=True)
    loss(prediction, target)


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_multi_scale_gmsd_loss_on_gpu(prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction = prediction.cuda()
    target = target.cuda()
    loss = MultiScaleGMSDLoss(chromatic=True)
    loss(prediction, target)


def test_multi_scale_gmsd_loss_backward(prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction.requires_grad_()
    loss_value = MultiScaleGMSDLoss(chromatic=True)(prediction, target)
    loss_value.backward()
    assert prediction.grad is not None, LEAF_VARIABLE_ERROR_MESSAGE


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_multi_scale_gmsd_loss_backward_on_gpu(prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction.requires_grad_()
    loss_value = MultiScaleGMSDLoss(chromatic=True)(prediction, target)
    loss_value.backward()
    assert prediction.grad is not None, LEAF_VARIABLE_ERROR_MESSAGE
    

def test_multi_scale_gmsd_zero_for_equal_tensors(prediction: torch.Tensor):
    loss = MultiScaleGMSDLoss()
    target = prediction.clone()
    measure = loss(prediction, target)
    assert measure.abs() <= 1e-6, f'MultiScaleGMSD for equal tensors must be 0, got {measure}'


def test_multi_scale_gmsd_loss_supports_different_data_ranges(
        prediction: torch.Tensor, target: torch.Tensor) -> None:
    prediction_255 = (prediction * 255).type(torch.uint8)
    target_255 = (target * 255).type(torch.uint8)
    loss = MultiScaleGMSDLoss()
    measure = loss(prediction, target)
    loss_255 = MultiScaleGMSDLoss(data_range=255)
    measure_255 = loss_255(prediction_255, target_255)
    diff = torch.abs(measure_255 - measure)
    assert diff <= 1e-4, f'Result for same tensor with different data_range should be the same, got {diff}'


def test_multi_scale_gmsd_supports_greyscale_tensors():
    loss = MultiScaleGMSDLoss()
    target = torch.ones(2, 1, 128, 128)
    prediction = torch.zeros(2, 1, 128, 128)
    loss(prediction, target)


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_multi_scale_gmsd_supports_greyscale_tensors_on_gpu():
    loss = MultiScaleGMSDLoss()
    target = torch.ones(2, 1, 128, 128).cuda()
    prediction = torch.zeros(2, 1, 128, 128).cuda()
    loss(prediction.cuda(), target.cuda())


def test_multi_scale_gmsd_fails_for_greyscale_tensors_chromatic_flag():
    loss = MultiScaleGMSDLoss(chromatic=True)
    target = torch.ones(2, 1, 128, 128)
    prediction = torch.zeros(2, 1, 128, 128)
    with pytest.raises(AssertionError):
        loss(prediction, target)


def test_multi_scale_gmsd_supports_custom_weights(
        prediction: torch.Tensor, target: torch.Tensor):
    loss = MultiScaleGMSDLoss(scale_weights=[3., 4., 2., 1., 2.])
    loss(prediction, target)


@pytest.mark.skipif(not torch.cuda.is_available(), reason='No need to run test if there is no GPU.')
def test_multi_scale_gmsd_supports_custom_weights_on_gpu(
        prediction: torch.Tensor, target: torch.Tensor):
    loss = MultiScaleGMSDLoss(scale_weights=[3., 4., 2., 1., 2.])
    loss(prediction, target)


def test_multi_scale_gmsd_raise_exception_for_small_images():
    target = torch.ones(3, 1, 32, 32)
    prediction = torch.zeros(3, 1, 32, 32)
    loss = MultiScaleGMSDLoss(scale_weights=[3., 4., 2., 1., 1.])
    with pytest.raises(ValueError):
        loss(prediction, target)



from piq.functional.base import ifftshift, get_meshgrid, similarity_map, gradient_map
from piq.functional.colour_conversion import rgb2lmn, rgb2xyz, xyz2lab, rgb2lab, rgb2yiq
from piq.functional.filters import hann_filter, scharr_filter, prewitt_filter, gaussian_filter
from piq.functional.layers import L2Pool2d


__all__ = [
    'ifftshift', 'get_meshgrid', 'similarity_map', 'gradient_map',
    'rgb2lmn', 'rgb2xyz', 'xyz2lab', 'rgb2lab', 'rgb2yiq',
    'hann_filter', 'scharr_filter', 'prewitt_filter', 'gaussian_filter',
    'L2Pool2d',
]
